#include "../common.h"
#include "../find.c"
#include "../CR0.c"

static void **taddr;                                        // table sys_call_table address

char* in4_ntoa( u32 ip ) {                                  // mapping IP to a string
   static char saddr[ MAX_ADDR_LEN ];
   sprintf( saddr, "%d.%d.%d.%d",
            ( ip >> 24 ) & 0xFF, ( ip >> 16 ) & 0xFF, 
            ( ip >> 8 ) & 0xFF, ( ip ) & 0xFF
          );
   return saddr;
}

static struct mutex mread,                                  // request wait mutex
                    mwrite;                                 // permission wait mutex

static char req_buff[ LEN_MSG ],                            // request buffer
            rep_buff = 1;                                   // reply buffer

// <linux/syscalls.h> :
asmlinkage long (*old_sys_socketcall) ( int call, unsigned long __user *args );
asmlinkage long (*old_sys_shutdown) ( int, int );

asmlinkage long new_sys_socketcall( int call, unsigned long __user *args ) {
#define PARMS 3
   /* <linux/socket.h> :
      struct sockaddr {
         sa_family_t     sa_family;      // address family, AF_xxx
         char            sa_data[14];    // 14 bytes of protocol address
   }; */
   static unsigned long a[ PARMS ];                         // accept() and connect() have the same number of parameters 3
   static struct sockaddr sa;
   // ----------- nested functions are a GCC extension ---------
   long get_addr( void ) {
      const unsigned int len = PARMS * sizeof( unsigned long );
      if( copy_from_user( a, args, len ) )
         return -EFAULT;
      if( copy_from_user( &sa, (struct sockaddr __user*)a[ 1 ], sizeof( struct sockaddr ) ) )
         return -EFAULT;
      return 0;
   }
   // ----------------------------------------------------------
   /* <linux/in.h> :
      struct sockaddr_in {
         __kernel_sa_family_t  sin_family;     // Address family
         __be16                sin_port;       // Port number
         struct in_addr        sin_addr;       // Internet address
            ... 
      };
      struct in_addr {
         __be32  s_addr;
      }; */
   /* <linux/byteorder/generic.h> :
      #define ___ntohl(x) __be32_to_cpu(x)
      #define ___ntohs(x) __be16_to_cpu(x)
      #define ntohl(x) ___ntohl(x)
      #define ntohs(x) ___ntohs(x) */
   long ret;
   if( SYS_ACCEPT == call ) {                               // accept() before syscall
      long err;
      if( ( err = get_addr() ) < 0 ) return err;
      if( AF_INET == sa.sa_family ) {                       // only IPv4
         struct sockaddr_in *usin = (struct sockaddr_in *)&sa;
         sprintf( req_buff, "accept: local_port=%d", ntohs( usin->sin_port ) );
         LOG( "%s\n", req_buff );
         mutex_unlock( &mread );
      }
   }
   if( SYS_CONNECT == call ) {                              // connect() before syscall
      long err;
      if( ( err = get_addr() ) < 0 ) return err;
      if( AF_INET == sa.sa_family ) {                       // only IPv4
         struct sockaddr_in *usin = (struct sockaddr_in *)&sa;
         sprintf( req_buff, "connect: remote_address=%s remote_port=%d",
                  in4_ntoa( ntohl( usin->sin_addr.s_addr ) ), ntohs( usin->sin_port ) );
         LOG( "%s\n", req_buff );
         mutex_unlock( &mread );
      }
   }
   ret = old_sys_socketcall( call, args );                  // retranslate to original sys_socketcall()
   if( SYS_ACCEPT == call ) {                               // accepr() after syscall
      long err;
      if( ( err = get_addr() ) < 0 ) return err;
      if( AF_INET == sa.sa_family ) {                       // only IPv4
         struct sockaddr_in *usin = (struct sockaddr_in *)&sa;
         sprintf( req_buff, "accept: remote_address=%s remote_port=%d",
                  in4_ntoa( ntohl( usin->sin_addr.s_addr ) ), ntohs( usin->sin_port ) );
         LOG( "%s\n", req_buff );
         mutex_unlock( &mread );
      }
   }
   return ret;
}
EXPORT_SYMBOL( new_sys_socketcall );

static ssize_t read( struct file *file, char *buf, size_t count, loff_t *ppos ) {
   int len = 0;
   int ret = mutex_lock_killable( &mread );
   if( ret != 0 ) return ret;
   len = strlen( req_buff );
   if( copy_to_user( buf, req_buff, len ) )
      return -EFAULT;
   return len;
}

static ssize_t write( struct file *file, const char *buf, size_t count, loff_t *ppos ) {
   int ret;
//, len = count < LEN_MSG ? count : LEN_MSG;
   ret = copy_from_user( &rep_buff, (void*)buf, sizeof( rep_buff ) );
   if( ret != 0 ) return ret;
   mutex_unlock( &mwrite );
LOG( "<---- write: %d\n", rep_buff );
   return sizeof( rep_buff );
}

static int open( struct inode *n, struct file *f ) {
   mutex_lock( &mread );
   mutex_lock( &mwrite );
   return 0;
}

static int release( struct inode *n, struct file *f ) {
   mutex_unlock( &mread );
   mutex_unlock( &mwrite );
   return 0;
}

static const struct file_operations fw_fops = {
   .owner   = THIS_MODULE,
   .open    = open,
   .release = release,
   .read    = read,
   .write   = write,
};

static struct miscdevice fw_dev = {
   MISC_DYNAMIC_MINOR, DEVNAME, &fw_fops
};

static int __init init( void ) {
   void *waddr;
   // ----------- nested functions are a GCC extension ---------
   int pos_in_table( const char *symbol ) {                 // position in sys_call_table (__NR_*)
      const int last = __NR_process_vm_writev;              // near last syscall in i386
      int n;
      waddr = find_sym( symbol );
      if( NULL == waddr ) return -1;
      for( n = 0; n <= last; n++ )
         if( taddr[ n ] == waddr ) break;
      return n <= last ? n : -1;
   }
   // --------------------------------------------------------
   void show_in_table( char *symb ) {                       // print info about symbol
      waddr = find_sym( symb );
      if( NULL == waddr ) {
         DEB( "symbol %s not found in kernel\n", symb );
      }
      else {
         int n = pos_in_table( symb );
         if( n > 0 ) {
            DEB( "symbol %s address = %p, position in sys_call_table = %d\n", symb, waddr, n );
         }
         else
            DEB( "symbol %s address = %p, not found in sys_call_table\n", symb, waddr );
      }
   }
   // --------------------------------------------------------
   int ret = misc_register( &fw_dev );                      // register /dev/policy
   if( ret ) {
      ERR( "unable to register device /dev/%s\n", DEVNAME );
      return ret;
   }
   LOG( "registered device /dev/%s\n", DEVNAME );
   mutex_init( &mread );
   mutex_init( &mwrite );
//   ipdeny = ntohl( deny != NULL ? in_aton( deny ) : in_aton( "0.0.0.0" ) );
//   LOG( "denied IP: %s\n", deny != NULL ? in4_ntoa( ipdeny ) : "no" );
//   if( port != 0 )
//      LOG( "denied TCP port: %d\n", port );
   if( NULL == ( taddr = find_sym( "sys_call_table" ) ) ) {
      ERR( "sys_call_table not found\n" ); 
      goto err;
   }
   DEB( "sys_call_table address = %p\n", taddr );
   old_sys_socketcall = (void*)taddr[ __NR_socketcall ];
   if( NULL == ( waddr = find_sym( "sys_socketcall" ) ) ) { // sys_socketcall not exported
      ERR( "sys_socketcall not found\n" ); 
      goto err;
   }
   if( old_sys_socketcall != waddr ) {                      // reinsurance!
      ERR( "Oooops! I don't understand: addresses not equal\n" );
      goto err;
   }
   if( NULL == ( waddr = find_sym( "sys_shutdown" ) ) ) {   // sys_shutdown not exported
      ERR( "sys_shutdown not found\n" ); 
      goto err;
   }
   old_sys_shutdown = waddr;
   if( debug ) show_cr0();
   rw_enable();
   taddr[ __NR_socketcall ] = new_sys_socketcall;
   if( debug ) show_cr0();
   rw_disable();
   if( debug ) show_cr0();
   LOG( "install new sys_socketcall handler: %p\n", &new_sys_socketcall );
   return 0;
err:
   mutex_destroy( &mread );
   mutex_destroy( &mwrite );
   misc_deregister( &fw_dev );                              // deregister /dev/policy
   return -EINVAL;
}

static void __exit exit( void ) {
   LOG( "sys_socketcall handler before unload: %p\n", (void*)taddr[ __NR_socketcall ] );
   rw_enable();
   taddr[ __NR_socketcall ] = old_sys_socketcall;
   rw_disable();
   LOG( "restore old sys_socketcall handler: %p\n", (void*)taddr[ __NR_socketcall ] );
   mutex_destroy( &mread );
   mutex_destroy( &mwrite );
   misc_deregister( &fw_dev );                              // deregister /dev/policy
   LOG( "deregistered device /dev/%s\n", DEVNAME );
   return;
}

module_init( init );
module_exit( exit );

MODULE_LICENSE( "GPL" );
MODULE_AUTHOR( "Oleg Tsiliuric <olej.tsil@gmail.com>" );
MODULE_VERSION( "2.2" );
